//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "FormatMonitor.h"
#include "SCSICmd.h"
#include "SCSIRunner.h"
#pragma package(smart_init)

int wait_for_unit(SCSICmd &cmd, int *progress, FormatMonitor *monitor) {
	unsigned char *sense = cmd.sense();
    unsigned char sensebuf[18];
    while (true) {
		cmd[0] = 0;	// TEST UNIT READY
		cmd[5] = 0;
		if (!cmd.transport ()) break;
		if ((sense[0]&0x70)==0) return 1;
		while (progress) {
			if (sense[15]&0x80) {
	    		*progress = sensebuf[16]<<8|sensebuf[17];
				break;
	    	}
		    cmd[0] = 0x03;	// REQUEST SENSE
		    cmd[4] = sizeof(sensebuf);
		    cmd[5] = 0;
		    if (cmd.transport(READ,sensebuf,sizeof(sensebuf))) return 1;
		    if (sensebuf[15]&0x80) {
		    	int data = sensebuf[16]<<8|sensebuf[17];
		    	if (data > (*progress)) *progress = data;
			}
            monitor->Check4Update();
		    break;
		}
    }
	return 0;
}
//---------------------------------------------------------------------------

//   Important: Methods and properties of objects in VCL can only be
//   used in a method called using Synchronize, for example:
//
//      Synchronize(UpdateCaption);
//
//   where UpdateCaption could look like:
//
//      void __fastcall FormatMonitor::UpdateCaption()
//      {
//        Form1->Caption = "Updated in a thread";
//      }
//---------------------------------------------------------------------------

__fastcall FormatMonitor::FormatMonitor(bool CreateSuspended, TMainFrame *frame)
    : TThread(CreateSuspended)
{
    this->MainFrame = frame;
    this->mmc_profile = 0;
}
//---------------------------------------------------------------------------

int __fastcall FormatMonitor::GetMMCProfile(SCSICmd &cmd) {
    int profile = 0;
    unsigned char dformat[256];
    cmd[0] = 0x46;		// GET CONFIGURATION
    cmd[8] = 8;
    cmd[9] = 0;
    if (!cmd.transport(READ, dformat,8)) profile = dformat[6]<<8|dformat[7];
    return profile;
}
//---------------------------------------------------------------------------

void __fastcall FormatMonitor::Execute()
{
    //---- Place thread code here ----
    blankformat = false;
    SCSICmd cmd;
    progress = 0;
    AnsiString dev = MainFrame->DeviceComboBox->Text;
    dev = dev.SubString(0, dev.Pos(":"));
    if (cmd.associate(dev.c_str())) {
        this->mmc_profile = GetMMCProfile(cmd);
        unsigned char dinfo[256];
        cmd[0] = 0x51;		// READ DISC INFORMATION
        cmd[8] = sizeof(dinfo);
        cmd[9] = 0;
        if (cmd.transport(READ,dinfo,sizeof(dinfo))) {
            this->MainFrame->ShowStatus("READ DISC INFORMATION FAILED");
            return;
        } else if (dinfo[2]&3) {
            this->MainFrame->ShowStatus("DISC IS NONE-BLANK MEDIA");
        } else {
            this->MainFrame->ShowStatus("DISC IS BLANK MEDIA");
            blankformat = true;
        }
        ::Sleep(5000);
        //-----------------------------------------------------------
    	unsigned char formats[256];
        cmd[0] = 0x23;		// READ FORMAT CAPACITIES
        cmd[8] = 4;
        cmd[9] = 0;
        if (cmd.transport(READ, formats,4)) {
            MainFrame->ShowStatus("READ FORMAT CAPACITIES FAILED");
            return;
        }
        int len = formats[3];
        if (len&7 || len<16) {
            MainFrame->ShowStatus("ALLOCATION LENGTH ISNOT SANE");
            return;
        }
    	int i = 0;
        cmd[0] = 0x23;		// READ FORMAT CAPACITIES
        cmd[7] = (4+len)>>8;	// now with real length...
        cmd[8] = (4+len)&0xFF;
        cmd[9] = 0;
        if (cmd.transport(READ, formats, 4+len)) {
            MainFrame->ShowStatus("READ FORMAT CAPACITIES FAILED");
            return;
        }
        if (len != formats[3]) {
            MainFrame->ShowStatus("PARAMETER LENGTH INCONSISTENCY");
            return;
        }
   	    for (i=8;i<len;i+=8) if ((formats [4+i+4]>>2) == 0x26) break;
        if (i==len) {
            MainFrame->ShowStatus("CANNOT LOCATE APPROPRIATE FORMAT DESCRIPTOR");
            return;
        }
        int capacity = 0;
        capacity |= formats[4+i+0], capacity <<= 8;
        capacity |= formats[4+i+1], capacity <<= 8;
        capacity |= formats[4+i+2], capacity <<= 8;
        capacity |= formats[4+i+3];
        int lead_out = 0;
        lead_out |= formats[4+0], lead_out <<= 8;
        lead_out |= formats[4+1], lead_out <<= 8;
        lead_out |= formats[4+2], lead_out <<= 8;
        lead_out |= formats[4+3];

        formats[i+0] = 0;		// "Reserved"
        formats[i+1] = 2;		// "IMMED" flag
        formats[i+2] = 0;		// "Descriptor Length" (MSB)
        formats[i+3] = 8;		// "Descriptor Length" (LSB)
        //-----------------------------------------------------------
		cmd[0] = 0x04;		// FORMAT UNIT
		cmd[1] = 0x11;		// "FmtData" and "Format Code"
		cmd[5] = 0;
		if (!cmd.transport(WRITE,formats+i,12)) {
            this->prefix = "榡ƺϺо";
            if (wait_for_unit(cmd, &progress, this)) return;
        }
        //-----------------------------------------------------------
        if (!blankformat) {
            cmd[0] = 0x5B;	// CLOSE TRACK/SESSION
	        cmd[1] = 1;		// "IMMED" flag on
	        cmd[2] = 0;		// "Stop De-Icing"
    	    cmd[9] = 0;
	        if (!cmd.transport()) {
                this->prefix = "gJϬqT(Stop DeIcing)";
	    	    if (wait_for_unit(cmd, &progress, this)) {
                    SCSIDevice::eject(dev.c_str(), 1);
                    return;
                }
            }
        }
        //-----------------------------------------------------------
		cmd[0] = 0x5B;	// CLOSE TRACK/SESSION
		cmd[1] = 1;		// "IMMED" flag on
		cmd[2] = 2;		// "Close Session"
		cmd[9] = 0;
		if (!cmd.transport()) {
            this->prefix = "gJϬqT(Close Session)";
            if (wait_for_unit(cmd, &progress, this)) {
                SCSIDevice::eject(dev.c_str(), 1);
                return;
            }
        }
        //-----------------------------------------------------------
    } else this->MainFrame->ShowStatus("Can't associate with device F:");
    ::Sleep(1000);
    SCSIDevice::eject(dev.c_str(), 1);
}
//---------------------------------------------------------------------------

void __fastcall FormatMonitor::Check4Update() {
    Synchronize(this->SetInfoStat);
}
//---------------------------------------------------------------------------

void __fastcall FormatMonitor::SetInfoStat() {
    double target = 988;
    if (blankformat) target = 6771;
    this->MainFrame->ShowStatus(AnsiString(prefix) + " : " + IntToStr(progress));
    this->MainFrame->CGauge->Progress = 100*progress/target;
}
//---------------------------------------------------------------------------
